import numpy as np
from typing import List


def get_tour_distance(
    Tour: List,
    dist_matrix: np.array
) -> float:
    """
    Given an array of indexes, calculate the distance through the
    distance matrix.

    Args:
        Tour (np.array): array of indexes
        dist_matrix (np.array): 2-d array of distance values

    Returns:
        float: total distance of the Tour.
    """
    Tour = Tour[0:-1]
    d = 0
    for i in range(len(Tour)):
        city_index = Tour[i]
        # The TSP always begin and end in the same city. When
        # in the last city in the tour go back to the start.
        try:
            next_city_index = Tour[i + 1]
        except IndexError:
            next_city_index = Tour[0]
        d += dist_matrix[city_index][next_city_index]
    return d



import pandas as pd
import numpy as np
import tsplib95

def distance(a, b):
    """Euclidean distance of two points.

    Args:
        a (numpy.array or pd.Series)
        b (numpy.array or pd.Series)

    Returns:
        [float]
    """
    return np.sqrt(np.sum(np.power((a-b), 2)))

def get_dist_matrix(path):
    """Given a csv of rows each containing
    index, x, y coordinates, return the distance matrix,
    for the group of points.
    """
    df = pd.read_csv(path,
                     sep=' ',
                     names=['x', 'y'])
    
    full_matrix = []
    for i in df.index:
        vec_distance = []
        for j in df.index:
            vec_distance.append(distance(df.iloc[i,].to_numpy(),
                                    df.iloc[j,].to_numpy()))
        vec_distance = np.array(vec_distance)
        full_matrix.append(vec_distance)
    full_matrix = np.array(full_matrix)
    return full_matrix

def tsplib_distance_matrix(tsplib_file: str) -> np.ndarray:
    """Distance matrix from a TSPLIB file
    Parameters
    ----------
    tsplib_file
        A string with the complete path of the TSPLIB file (or just its name if
        it is the in current path)
    Returns
    -------
    distance_matrix
        A ND-array with the equivalent distance matrix of the input file
    Notes
    -----
    This function can handle any file supported by the `tsplib95` lib.
    """

    tsp_problem = tsplib95.load(tsplib_file)
    distance_matrix_flattened = np.array([
        tsp_problem.get_weight(*edge) for edge in tsp_problem.get_edges()
    ])
    distance_matrix = np.reshape(
        distance_matrix_flattened,
        (tsp_problem.dimension, tsp_problem.dimension),
    )

    # Some problems with EXPLICIT matrix have a large number in the distance
    # from a node to itself, which makes no sense for our problems. Thus,
    # always ensure a diagonal filled with zeros
    np.fill_diagonal(distance_matrix, 0)
    return distance_matrix


"""Pure GRASP solver"""
from random import sample
from timeit import default_timer
from typing import List, Optional, Tuple

import numpy as np
from tsp_heuristics.heuristics.local_search import local_search_algo
from tsp_heuristics.heuristics.utils import get_tour_distance as compute_permutation_distance

# Testing
from tsp_heuristics.utils.readfile import tsplib_distance_matrix
from tsp_heuristics.heuristics.nn import nn_algo

def solve_tsp_grasp(
    distance_matrix: np.ndarray,
    start_position: int = 0,
    alpha: float = 0.5,
    perturbation_scheme: str = "two_opt",
    max_iterations: int = 1, 
    max_processing_time: Optional[float] = None,
    log_file: Optional[str] = None,
) -> Tuple[List, float, float]:
    """Solve a TSP problem with a GRASP heuristic
    """
    max_processing_time = max_processing_time or np.inf

    tic = default_timer()
    best_Tour = setup(distance_matrix, None)[0]
    i = 0
    while i <= max_iterations:
        intial_Tour = constructive_phase(distance_matrix, alpha, start_position)
        optimized_Tour = local_search_algo(intial_Tour, distance_matrix,method='first improvement')[0]
        
        f_best_tour = compute_permutation_distance(best_Tour, distance_matrix)
        f_optimized_tour = compute_permutation_distance(optimized_Tour, distance_matrix)
        
        if f_best_tour > f_optimized_tour:
            best_Tour = optimized_Tour
        
        if default_timer() - tic > max_processing_time:
            i = max_iterations + 1
            break
        
        i += 1
    return best_Tour, compute_permutation_distance(best_Tour, distance_matrix), (default_timer() - tic)


def constructive_phase(
    distance_matrix: np.ndarray,
    alpha: float,
    start: int = 0
) -> List:    
    Tour = [start]
    
    for _ in range(distance_matrix.shape[0] - 1):
        min_index = get_maxmin_index_from_row(distance_matrix, Tour[-1], Tour, 'min')
        max_index = get_maxmin_index_from_row(distance_matrix, Tour[-1], Tour, 'max')
        
        f_min = distance_matrix[Tour[-1]][min_index]
        f_max = distance_matrix[Tour[-1]][max_index]
        
        # List of Restrict Candidates = LRC
        LRC_index = np.array(range(len(distance_matrix[Tour[-1]])))
        
        LRC_condition = distance_matrix[Tour[-1]] <= f_min + alpha*(f_max - f_min)
        LRC_condition[Tour[-1]] = False
        LRC_index = LRC_index[LRC_condition]
        
        new_city_index = np.random.choice(LRC_index, 1, replace=False)[0]
        Tour.append(new_city_index)

    return Tour


def get_maxmin_index_from_row(
    distance_matrix: np.ndarray,
    row: int,
    previous_indexes: List,
    type: str,
    )-> int:
    """Get the minimum/maximum element in the adjusted row array from a distance matrix.
    We adjust the row array in order to never get the "previous_indexes" list of indexes.
    """
    distance_matrix = distance_matrix.copy()
    arr = distance_matrix[row].astype(float)
    
    aux_list = range(arr.shape[0])
    aux_list_2 = []
    for i in aux_list:
        if i in previous_indexes:
            aux_list_2.append(True)
        else:
            aux_list_2.append(False)
    previous_indexes_bool = aux_list_2
    
    if type == 'max':
        arr[previous_indexes_bool] = -1
        target_index = np.argmax(arr)
    if type == 'min':
        arr[previous_indexes_bool] = np.Inf
        target_index = np.argmin(arr)
    
    return target_index


def setup(
    distance_matrix: np.ndarray, x0: Optional[List] = None
) -> Tuple[List[int], float]:
    """Return initial solution and its objective value

    Parameters
    ----------
    distance_matrix
        Distance matrix of shape (n x n) with the (i, j) entry indicating the
        distance from node i to j

    x0
        Permutation of nodes from 0 to n - 1 indicating the starting solution.
        If not provided, a random list is created.

    Returns
    -------
    x0
        Permutation with initial solution. If ``x0`` was provided, it is the
        same list

    fx0
        Objective value of x0
    """

    if not x0:
        n = distance_matrix.shape[0]  # number of nodes
        x0 = [0] + sample(range(1, n), n - 1)  # ensure 0 is the first node

    fx0 = compute_permutation_distance(x0, distance_matrix)
    return x0, fx0


if __name__ == '__main__':
    tsplib_file = "tests/test_data/a280.tsp"
    distance_matrix = tsplib_distance_matrix(tsplib_file)

    print(solve_tsp_grasp(distance_matrix))
    
    initial_tour = nn_algo(distance_matrix)[0]
    print(local_search_algo(initial_tour, distance_matrix, method='first improvement'))
    
    
        


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tsp_heuristics', 'tsp_heuristics.heuristics', 'tsp_heuristics.utils']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4.2,<4.0.0',
 'numpy>=1.21.1,<2.0.0',
 'pandas>=1.3.0,<2.0.0',
 'sympy>=1.8,<2.0',
 'tsplib95>=0.7.1,<0.8.0']

setup_kwargs = {
    'name': 'tsp-heuristics',
    'version': '0.2.0',
    'description': 'Meta-heuristics for solving the TSP',
    'long_description': None,
    'author': 'Adriel Martins',
    'author_email': 'adrielfalcao@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import numpy as np
from random import sample
from tsp_heuristics.heuristics.utils import get_tour_distance

from typing import Generator, List, Tuple
from time import time


def two_opt_gen(
    Tour: List
) -> Generator:
    """
    Generates a neighborhood of 2-opt in the Tour, which means deleting
    two edges from the Tour and swap it. It creates in a random order all the
    neighborhood.

    Args:
        Tour (List)

    Yields:
        Generator: List of neighborhood from the 2-opt method.
    """
    x = Tour
    n = len(x)
    i_range = range(2, n)
    for i in sample(i_range, len(i_range)):
        j_range = range(i + 1, n + 1)
        for j in sample(j_range, len(j_range)):
            xn = x
            xn = xn[:i - 1] + list(reversed(xn[i - 1:j])) + xn[j:]
            yield xn


def local_search_algo(
    Tour: List,
    dist_matrix: np.array,
    method: str = 'best improvement'
) -> Tuple[List, float, float]:
    """
    In summary, search through 3 available methods the best solution in the
    neighboring initial solution or Tour.

    Args:
        Tour (List)
        dist_matrix (np.array)
        method (str, optional): One of the 3 available methods:
        'first improvement', 'best improvement' and 'random'.
        Defaults to 'best improvement'.

    Returns:
        Tuple[List, float]: Tour and time to execute the algorithm.
    """
    t0 = time()
    
    early_stop = False if method == 'best improvement' else True
    improvement = True
    while improvement:
        improvement = False
        best_Tour_distance = get_tour_distance(Tour, dist_matrix)
        for n in two_opt_gen(Tour):
            n_tour_distance = get_tour_distance(n, dist_matrix)
            if n_tour_distance < best_Tour_distance:
                improvement = True
                Tour = n
                best_Tour_distance = n_tour_distance
                if early_stop:
                    break
    
    return Tour, get_tour_distance(Tour, dist_matrix), (time() - t0)




import numpy as np
from time import time
from typing import List, Tuple
from tsp_heuristics.heuristics.utils import get_tour_distance

def nn_algo(
    dist_matrix: np.array,
    start: int = 0
) -> Tuple[List, float]:
    """
    From a start city index, get an Tour according to the Nearest Neighbor
    algorithm from the collection of the cities indexes.
    
    Args:
        dist_matrix (np.array)
        start (int, optional): The first city that we will begin the Tour and
        eventually return. Defaults to 0.

    Returns:
        np.array: Array of indexes representing the city Tour.
        float: Time to complete the algorithm.
    """
    t0 = time()
    Tour = [start]
    
    dist_matrix = dist_matrix.astype(float)
    # Making the distance to go to the same
    # city impossible.
    for i in range(dist_matrix.shape[0]):
        dist_matrix[i][i] = np.Inf
    
    for _ in range(dist_matrix.shape[0] - 1):
        # Finding the best next city.
        min_index = np.argmin(dist_matrix[Tour[-1]])
        # Making sure that we won't revisit
        # the same city.
        for t in Tour:
            dist_matrix[min_index][t] = np.Inf
            dist_matrix[t][min_index] = np.Inf
        Tour.append(min_index)
    
    return Tour, get_tour_distance(Tour, dist_matrix), (time() - t0)
